//
//  main.cpp
//  DUO_Light_emulator
//
//  Created by Jack Eisenmann on 7/7/14.
//  Copyright (c) 2014 Jack Eisenmann. All rights reserved.
//

#include <iostream>
#include <time.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <math.h>
#include <sstream>

#ifdef __APPLE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/glu.h>
#include <GL/gl.h>
#endif

#define byte unsigned char
#define true 1
#define false 0
#define long int

#define VIDEO_CHARACTER_WIDTH 17
#define VIDEO_CHARACTER_HEIGHT 14
#define PIXEL_SIZE 4
#define VIEWPORT_WIDTH (VIDEO_CHARACTER_WIDTH * 6 * PIXEL_SIZE)
#define VIEWPORT_HEIGHT (VIDEO_CHARACTER_HEIGHT * 8 * PIXEL_SIZE)
#define FRAME_DELAY 50

#define MAXIMUM_FILE_NAME_LENGTH 20

#define KEY_IS_HELD_BUFFER_SIZE 8

#define PROMPT_TEXT_CURSOR_CHARACTER '_'
#define PROMPT_SELECTION_CURSOR_CHARACTER '>'

#define OPCODE_NAME_MAXIMUM_SIZE 4

#define EXTERNAL_SRAM_SIZE ((long)256 * (long)256)
#define HEAP_START_ADDRESS EXTERNAL_SRAM_SIZE

#define PROGRAM_COUNTER_ADDRESS EXTERNAL_SRAM_SIZE
#define GLOBAL_SCOPE_ADDRESS_ADDRESS (EXTERNAL_SRAM_SIZE + 2)
#define LOCAL_SCOPE_ADDRESS_ADDRESS (EXTERNAL_SRAM_SIZE + 4)

#define SRAM_BUFFER_SIZE 100

#define GAME_OF_LIFE_DEAD_CELL ' '
#define GAME_OF_LIFE_LIVE_CELL 1
#define MANDELBROT_START_CHARACTER 11

using namespace std;

const byte FONT_DATA[] = {
	0, 0, 0, 0, 0, 0, 0, 0,
    // Solid block.
	31, 31, 31, 31, 31, 31, 31, 0,
    // Semi solid block.
    31, 21, 27, 21, 27, 21, 31, 0,
    // Player.
    31, 21, 31, 17, 31, 9, 27, 0,
    // Flower.
	27, 21, 14, 21, 31, 4, 4, 0,
    // Spikes.
	0, 21, 14, 31, 14, 21, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
    // Newline.
	0, 0, 0, 0, 0, 0, 0, 0,
    // Mandelbrot.
	0, 0, 0, 0, 0, 0, 0, 0,
	3, 3, 3, 0, 0, 0, 0, 0,
	24, 24, 24, 0, 0, 0, 0, 0,
	27, 27, 27, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 3, 3, 3, 0,
	3, 3, 3, 0, 3, 3, 3, 0,
	24, 24, 24, 0, 3, 3, 3, 0,
	27, 27, 27, 0, 3, 3, 3, 0,
	0, 0, 0, 0, 24, 24, 24, 0,
	3, 3, 3, 0, 24, 24, 24, 0,
	24, 24, 24, 0, 24, 24, 24, 0,
	27, 27, 27, 0, 24, 24, 24, 0,
	0, 0, 0, 0, 27, 27, 27, 0,
	3, 3, 3, 0, 27, 27, 27, 0,
	24, 24, 24, 0, 27, 27, 27, 0,
	27, 27, 27, 0, 27, 27, 27, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	// Space.
	0, 0, 0, 0, 0, 0, 0, 0,
	4, 4, 4, 4, 4, 0, 4, 0,
	10, 10, 10, 0, 0, 0, 0, 0,
	10, 10, 31, 10, 31, 10, 10, 0,
	4, 30, 5, 14, 20, 15, 4, 0,
	// Percent.
	17, 16, 8, 4, 2, 1, 17, 0,
	2, 5, 5, 22, 9, 9, 22, 0,
	4, 4, 4, 0, 0, 0, 0, 0,
	8, 4, 2, 2, 2, 4, 8, 0,
	2, 4, 8, 8, 8, 4, 2, 0,
	// Asterisk.
	10, 4, 10, 0, 0, 0, 0, 0,
	0, 4, 4, 31, 4, 4, 0, 0,
	0, 0, 0, 0, 4, 4, 4, 0,
	0, 0, 0, 31, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 4, 0,
	// Slash.
	16, 16, 8, 4, 2, 1, 1, 0,
	14, 17, 25, 21, 19, 17, 14, 0,
	4, 6, 5, 4, 4, 4, 31, 0,
	15, 16, 16, 14, 1, 1, 31, 0,
	15, 16, 16, 14, 16, 16, 15, 0,
	// 4.
	17, 17, 17, 31, 16, 16, 16, 0,
	31, 1, 1, 15, 16, 16, 15, 0,
	30, 1, 1, 15, 17, 17, 14, 0,
	31, 16, 8, 4, 4, 4, 4, 0,
	14, 17, 17, 14, 17, 17, 14, 0,
	// 9.
	14, 17, 17, 30, 16, 16, 15, 0,
	0, 4, 0, 0, 0, 4, 0, 0,
	0, 4, 0, 0, 4, 4, 4, 0,
	8, 4, 2, 1, 2, 4, 8, 0,
	0, 0, 31, 0, 31, 0, 0, 0,
	// Greater than.
	2, 4, 8, 16, 8, 4, 2, 0,
	15, 16, 16, 12, 4, 0, 4, 0,
	14, 17, 29, 29, 29, 1, 30, 0,
	14, 17, 17, 31, 17, 17, 17, 0,
	15, 17, 17, 15, 17, 17, 15, 0,
	// C.
	30, 1, 1, 1, 1, 1, 30, 0,
	15, 17, 17, 17, 17, 17, 15, 0,
	31, 1, 1, 15, 1, 1, 31, 0,
	31, 1, 1, 15, 1, 1, 1, 0,
	30, 1, 1, 25, 17, 17, 14, 0,
	// H.
	17, 17, 17, 31, 17, 17, 17, 0,
	31, 4, 4, 4, 4, 4, 31, 0,
	28, 16, 16, 16, 16, 16, 15, 0,
	17, 9, 5, 3, 5, 9, 17, 0,
	1, 1, 1, 1, 1, 1, 31, 0,
	// M.
	17, 27, 21, 21, 17, 17, 17, 0,
	17, 19, 21, 25, 17, 17, 17, 0,
	14, 17, 17, 17, 17, 17, 14, 0,
	15, 17, 17, 15, 1, 1, 1, 0,
	14, 17, 17, 17, 21, 9, 22, 0,
	// R.
	15, 17, 17, 15, 5, 9, 17, 0,
	30, 1, 1, 14, 16, 16, 15, 0,
	31, 4, 4, 4, 4, 4, 4, 0,
	17, 17, 17, 17, 17, 17, 14, 0,
	17, 17, 17, 10, 10, 4, 4, 0,
	// W.
	17, 17, 17, 21, 21, 27, 17, 0,
	17, 17, 10, 4, 10, 17, 17, 0,
	17, 17, 10, 4, 4, 4, 4, 0,
	31, 16, 8, 4, 2, 1, 31, 0,
	14, 2, 2, 2, 2, 2, 14, 0,
	// Backslash.
	1, 1, 2, 4, 8, 16, 16, 0,
	14, 8, 8, 8, 8, 8, 14, 0,
	4, 10, 17, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 31, 0,
	2, 4, 8, 0, 0, 0, 0, 0,
	// a.
	0, 0, 15, 16, 30, 17, 30, 0,
	1, 1, 15, 17, 17, 17, 15, 0,
	0, 0, 30, 1, 1, 1, 30, 0,
	16, 16, 30, 17, 17, 17, 30, 0,
	0, 0, 14, 17, 31, 1, 30, 0,
	// f.
	24, 4, 31, 4, 4, 4, 4, 0,
	0, 0, 30, 17, 30, 16, 15, 0,
	1, 1, 15, 17, 17, 17, 17, 0,
	4, 0, 4, 4, 4, 4, 4, 0,
	16, 0, 16, 16, 16, 16, 15, 0,
	// k.
	1, 1, 17, 9, 7, 9, 17, 0,
	4, 4, 4, 4, 4, 4, 4, 0,
	0, 0, 15, 21, 21, 21, 21, 0,
	0, 0, 15, 17, 17, 17, 17, 0,
	0, 0, 14, 17, 17, 17, 14, 0,
	// p.
	0, 0, 15, 17, 15, 1, 1, 0,
	0, 0, 30, 17, 30, 16, 16, 0,
	0, 0, 31, 1, 1, 1, 1, 0,
	0, 0, 30, 1, 14, 16, 15, 0,
	4, 4, 31, 4, 4, 4, 4, 0,
	// u.
	0, 0, 17, 17, 17, 17, 30, 0,
	0, 0, 17, 17, 10, 10, 4, 0,
	0, 0, 21, 21, 21, 21, 10, 0,
	0, 0, 17, 10, 4, 10, 17, 0,
	0, 0, 17, 17, 30, 16, 15, 0,
	// z.
	0, 0, 31, 8, 4, 2, 31, 0,
	12, 2, 2, 1, 2, 2, 12, 0,
	4, 4, 4, 0, 4, 4, 4, 0,
	6, 8, 8, 16, 8, 8, 6, 0,
	0, 0, 18, 21, 9, 0, 0, 0
};

const byte KEYBOARD_CODE_DATA[] = {
	0x1C, 'a', 'A',
	0x32, 'b', 'B',
	0x21, 'c', 'C',
	0x23, 'd', 'D',
	0x24, 'e', 'E',
	0x2B, 'f', 'F',
	0x34, 'g', 'G',
	0x33, 'h', 'H',
	0x43, 'i', 'I',
	0x3B, 'j', 'J',
	0x42, 'k', 'K',
	0x4B, 'l', 'L',
	0x3A, 'm', 'M',
	0x31, 'n', 'N',
	0x44, 'o', 'O',
	0x4D, 'p', 'P',
	0x15, 'q', 'Q',
	0x2D, 'r', 'R',
	0x1B, 's', 'S',
	0x2C, 't', 'T',
	0x3C, 'u', 'U',
	0x2A, 'v', 'V',
	0x1D, 'w', 'W',
	0x22, 'x', 'X',
	0x35, 'y', 'Y',
	0x1A, 'z', 'Z',
	0x45, '0', ')',
	0x16, '1', '!',
	0x1E, '2', '@',
	0x26, '3', '#',
	0x25, '4', '$',
	0x2E, '5', '%',
	0x36, '6', '^',
	0x3D, '7', '&',
	0x3E, '8', '*',
	0x46, '9', '(',
	0x0E, '`', '~',
	0x4E, '-', '_',
	0x55, '=', '+',
	0x5D, '\\', '|',
	0x29, ' ', ' ',
	0x54, '[', '{',
	0x5B, ']', '}',
	0x4C, ';', ':',
	0x52, '\'', '"',
	0x41, ',', '<',
	0x49, '.', '>',
	0x4A, '/', '?',
};

const byte COMMAND_ARGUMENT_AMOUNT_LIST[] = {
    // WRT.
    3, 3, 5, 5,
    // OR.
    4, 4, 4,
    // BNT.
    3, 4, 4, 4,
    // ADD.
    4, 4, 4, 4,
    // INC.
    2, 2, 2, 2,
    // FNC.
    1, 0,
    // FIL.
    5, 6,
    // CII.
    4, 4,
    // MUL.
    4, 4, 4, 4, 4,
    // ADDF.
    3, 3, 3, 3,
    // CIF.
    3, 3, 3, 3,
    // CIT.
    3, 3, 2, 3, 3, 2,
    // WRTT.
    2, 3, 4, 3, 5, 2,
    // SDL.
    3, 4,
    // RAN.
    3, 2,
    // TWR.
    2, 2, 2,
    // DCL.
    0, 4, 6,
    // CKT.
    3, 2,
    // KIP.
    2, 1, 3, 4, 2,
    // XSM.
    2, 2, 2, 2,
    // RNF.
    2, 3, 2, 2, 4, 2, 2, 2, 3, 1, 1, 0,
    // CTO.
    2, 2, 2,
    // ALLO.
    3, 1,
    // DSC.
    2, 2, 2,
    // GMLF.
    5, 6,
    // POW.
    3, 3
};

const byte COMMAND_OPCODE_NAME_LIST[] = "WRT NOT WRZ WNZ OR  AND XOR BNT GRE GRESEQU ADD SUB BSL BSR INC DEC SLO SRO FNC RET FIL FND CII CIISMUL MULSDIV DIVSMOD ADDFSUBFMULFDIVFCIF CIFSCFI CFISCIT CITSCFT CTI CTISCTF WRTTEQUTFNDTLENTSUBTCONTSDL SDS RAN SIN TWR TRD TSL DCL DWR DRD CKT CTK KIP PKY PTX PSL PFL XSM XDR XDW XAR RNF RGAIRGANRGAARCR RRD RWR RGN RGS RDL RRN QUITCTO CIH CHI ALLOFREEDSC DDX DDY GMLFMNDLPOW LOG ";

const byte START_FILE_NAME[] = "start.dlbc";

const byte BOOT_MESSAGE_1[] = "No SD card.";
const byte BOOT_MESSAGE_2[] = "Found SD card.";
const byte BOOT_MESSAGE_3[] = "Should be FAT16.";
const byte BOOT_MESSAGE_4[] = "Correct format.";
const byte BOOT_MESSAGE_5[] = "Booting.";
const byte BOOT_MESSAGE_6[] = "No start file.";

const byte SD_CARD_DIRECTORY_PATH[] = "./SD card";

byte videoBuffer[VIDEO_CHARACTER_WIDTH * VIDEO_CHARACTER_HEIGHT];

byte externalSram[EXTERNAL_SRAM_SIZE];
// 0 = Closed
// 1 = Reading
// 2 = Writing
byte externalSramState = 0;
long externalSramAddress;

byte bootStage = 0;
unsigned long timeMilliseconds = 0;
byte isSleeping = false;
unsigned long wakeTimeMilliseconds;

DIR *sDDirectory;
byte fileNameList[200 * MAXIMUM_FILE_NAME_LENGTH];
short numberOfFiles;

unsigned short keyPressBuffer[8];
byte keyPressBufferEndIndex = 0;
unsigned short keyIsHeldBuffer[KEY_IS_HELD_BUFFER_SIZE];

// 0 = executing commands
// 1 = prompting key
// 2 = prompting text
// 3 = prompting selection
// 4 = prompting file
byte runtimeMode;
unsigned short promptIndex;
unsigned short promptIndex2;
unsigned short promptAmount;
unsigned short promptSize;
unsigned short promptVideoBufferIndex;
long promptSourceAddress;
long promptDestinationAddress;

unsigned short programCounter;
unsigned short globalScopeAddress;
unsigned short localScopeAddress;
byte addressSize;

byte pgm_read_byte(const byte *source)
{
    return *source;
}

void readTextFromProgMem(byte *destination, const byte *source)
{
	short index = 0;
	while (true)
	{
		byte tempCharacter = pgm_read_byte((byte *)source + index);
		index += 1;
		*destination = tempCharacter;
		destination += 1;
		if (tempCharacter == 0)
		{
			return;
		}
	}
}

void displayCharacter(byte posX, byte posY)
{
    byte tempCharacter = videoBuffer[(short)posX + ((short)posY) * VIDEO_CHARACTER_WIDTH];
    byte *tempPointer = ((byte *)FONT_DATA) + ((short)tempCharacter) * 8;
    byte tempOffsetY = 0;
    while (tempOffsetY < 7)
    {
        byte tempData = *tempPointer;
        tempPointer += 1;
        byte tempMask = 0x01;
        byte tempOffsetX = 0;
        while (tempOffsetX < 5)
        {
            if (tempData & tempMask)
            {
                short tempPosX = (((int)posX) * 6 + (short)tempOffsetX) * PIXEL_SIZE;
                short tempPosY = VIEWPORT_HEIGHT - (((short)posY) * 8 + (short)tempOffsetY) * PIXEL_SIZE;
                glRectf(tempPosX, tempPosY, tempPosX + PIXEL_SIZE, tempPosY - PIXEL_SIZE);
            }
            tempMask <<= 1;
            tempOffsetX += 1;
        }
        tempOffsetY += 1;
    }
}

void display()
{
    glClear(GL_COLOR_BUFFER_BIT);
    
    glColor3f(1, 1, 1);
    
    byte tempPosY = 0;
    while (tempPosY < VIDEO_CHARACTER_HEIGHT)
    {
        byte tempPosX = 0;
        while (tempPosX < VIDEO_CHARACTER_WIDTH)
        {
            displayCharacter(tempPosX, tempPosY);
            tempPosX += 1;
        }
        tempPosY += 1;
    }
    
    glFlush();
}

void reshape(int width, int height)
{
    glViewport(0, 0, VIEWPORT_WIDTH, VIEWPORT_HEIGHT);
    
    glMatrixMode(GL_PROJECTION);
    
    glLoadIdentity();
    
    gluOrtho2D(0, VIEWPORT_WIDTH, 0, VIEWPORT_HEIGHT);
    
    glMatrixMode(GL_MODELVIEW);
}

// Shift key is held: keyIsHeld(0x0012) || keyIsHeld(0x0059)
byte convertKeyToCharacter(unsigned short key, byte shiftKeyIsHeld)
{
    if (key & 0xFF00)
    {
        return 0;
    }
	short index = 0;
	while (index < sizeof(KEYBOARD_CODE_DATA))
	{
		byte tempKeyboardCode = pgm_read_byte(KEYBOARD_CODE_DATA + index);
		if (tempKeyboardCode == key)
		{
			if (shiftKeyIsHeld)
			{
				return pgm_read_byte(KEYBOARD_CODE_DATA + index + 2);
			} else {
				return pgm_read_byte(KEYBOARD_CODE_DATA + index + 1);
			}
		}
		index += 3;
	}
	return 0;
}

unsigned short convertCharacterToKey(byte character)
{
    short index = 0;
	while (index < sizeof(KEYBOARD_CODE_DATA))
	{
        byte tempCharacter1 = pgm_read_byte(KEYBOARD_CODE_DATA + index + 1);
        byte tempCharacter2 = pgm_read_byte(KEYBOARD_CODE_DATA + index + 2);
		if (character == tempCharacter1 || character == tempCharacter2)
        {
            return pgm_read_byte(KEYBOARD_CODE_DATA + index);
		}
		index += 3;
	}
    return 0;
}

unsigned short convertGlutKeyToPS2Key(byte key)
{
    unsigned short tempKey = convertCharacterToKey(key);
    if (tempKey)
    {
        return tempKey;
    }
    switch (key)
    {
        // Backspace.
        case 127:
        {
            return 0x66;
        }
        // Return.
        case 13:
        {
            return 0x5A;
        }
        // Escape.
        case 27:
        {
            return 0x76;
        }
        default:
        {
            break;
        }
    }
    return 0;
}

unsigned short convertGlutSpecialToPS2Key(int special)
{
    switch (special)
    {
        case GLUT_KEY_LEFT:
        {
            return 0xE06B;
        }
        case GLUT_KEY_RIGHT:
        {
            return 0xE074;
        }
        case GLUT_KEY_UP:
        {
            return 0xE075;
        }
        case GLUT_KEY_DOWN:
        {
            return 0xE072;
        }
        default:
        {
            break;
        }
    }
    return 0;
}

byte findHeldKey(unsigned short key)
{
	byte index = 0;
	while (index < KEY_IS_HELD_BUFFER_SIZE)
	{
		unsigned short tempKey = keyIsHeldBuffer[index];
		if (tempKey == key)
		{
			return index;
		}
		index += 1;
	}
	return 255;
}

byte keyIsHeld(unsigned short key)
{
	return (findHeldKey(key) != 255);
}

unsigned short getNextKeyPress()
{
    if (keyPressBufferEndIndex == 0)
    {
        return 0;
    }
    unsigned short output = keyPressBuffer[0];
    byte index = 1;
    while (index < keyPressBufferEndIndex)
    {
        keyPressBuffer[index - 1] = keyPressBuffer[index];
        index += 1;
    }
    keyPressBufferEndIndex -= 1;
    return output;
}

void registerKeyPress(unsigned short key)
{
    if (keyPressBufferEndIndex < sizeof(keyPressBuffer) / sizeof(unsigned short))
    {
        keyPressBuffer[keyPressBufferEndIndex] = key;
        keyPressBufferEndIndex += 1;
    }
    if (!keyIsHeld(key))
    {
        byte index = findHeldKey(0);
        keyIsHeldBuffer[index] = key;
    }
}

void registerKeyRelease(unsigned short key)
{
    byte index = findHeldKey(key);
    if (index != 255)
    {
        keyIsHeldBuffer[index] = 0;
    }
}

void updateShiftKey()
{
    if (!keyIsHeld(0x0012))
    {
        if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
        {
            registerKeyPress(0x0012);
        }
    }
    if (keyIsHeld(0x0012))
    {
        if (glutGetModifiers() != GLUT_ACTIVE_SHIFT)
        {
            registerKeyRelease(0x0012);
        }
    }
}

void keyboardEvent(byte key, int x, int y)
{
    updateShiftKey();
    unsigned short tempKey = convertGlutKeyToPS2Key(key);
    registerKeyPress(tempKey);
}

void keyboardUpEvent(byte key, int x, int y)
{
    updateShiftKey();
    unsigned short tempKey = convertGlutKeyToPS2Key(key);
    registerKeyRelease(tempKey);
}

void specialEvent(int special, int x, int y)
{
    unsigned short tempKey = convertGlutSpecialToPS2Key(special);
    registerKeyPress(tempKey);
}

void specialUpEvent(int special, int x, int y)
{
    unsigned short tempKey = convertGlutSpecialToPS2Key(special);
    registerKeyRelease(tempKey);
}

void copyData(byte *destination, unsigned short amount, byte *source)
{
	while (amount > 0)
	{
		byte tempCharacter = *source;
		*destination = tempCharacter;
		source += 1;
		destination += 1;
        amount -= 1;
	}
}

void copyText(byte *destination, byte *source)
{
	while (true)
	{
		byte tempCharacter = *source;
		*destination = tempCharacter;
		if (!tempCharacter)
		{
			break;
		}
		source += 1;
		destination += 1;
	}
}

byte equalText(byte *text1, byte *text2)
{
	while (true)
	{
		byte tempCharacter1 = *text1;
        byte tempCharacter2 = *text2;
        if (tempCharacter1 != tempCharacter2)
        {
            return false;
        }
		if (!tempCharacter1)
		{
			return true;
		}
		text1 += 1;
        text2 += 1;
	}
}

void convertLongToText(byte *destination, long number)
{
    ostringstream tempStream;
    tempStream << number;
    string tempString = tempStream.str();
    byte *tempText = (byte *)(tempString.c_str());
    copyText(destination, tempText);
}

void convertFloatToText(byte *destination, float number)
{
    ostringstream tempStream;
    tempStream << number;
    string tempString = tempStream.str();
    byte *tempText = (byte *)(tempString.c_str());
    copyText(destination, tempText);
}

long convertTextToLong(byte *text)
{
    return atoi((char *)text);
}

float convertTextToFloat(byte *text)
{
    return atof((char *)text);
}

byte convertTextToOpcode(byte *text)
{
    byte output = 0;
    unsigned short index = 0;
    while (index < sizeof(COMMAND_OPCODE_NAME_LIST))
    {
        byte textIsEqual = false;
        byte tempOffset = 0;
        while (tempOffset < OPCODE_NAME_MAXIMUM_SIZE + 1)
        {
            byte tempCharacter1 = *(text + tempOffset);
            byte tempCharacter2 = pgm_read_byte(COMMAND_OPCODE_NAME_LIST + index + tempOffset);
            byte tempCondition1 = (tempCharacter1 == 0);
            byte tempCondition2 = (tempCharacter2 == ' ' || tempOffset > OPCODE_NAME_MAXIMUM_SIZE - 1);
            if (tempCondition1 && tempCondition2)
            {
                textIsEqual = true;
            }
            if (tempCondition1 || tempCondition2)
            {
                break;
            }
            if (tempCharacter1 != tempCharacter2)
            {
                break;
            }
            tempOffset += 1;
        }
        if (textIsEqual)
        {
            return output;
        }
        output += 1;
        index += OPCODE_NAME_MAXIMUM_SIZE;
    }
    return 255;
}

unsigned long getTimeMilliseconds()
{
	return timeMilliseconds;
}

void setTimeMilliseconds(unsigned long number)
{
    timeMilliseconds = number;
}

void sleepMilliseconds(unsigned long delay)
{
    isSleeping = true;
	wakeTimeMilliseconds = getTimeMilliseconds() + delay;
}

void clearVideoBuffer()
{
	short index = 0;
	while (index < sizeof(videoBuffer))
	{
		videoBuffer[index] = ' ';
		index += 1;
	}
}

void drawText(byte posX, byte posY, byte *text)
{
	byte *videoBufferPointer = videoBuffer + (short)posX + (short)posY * VIDEO_CHARACTER_WIDTH;
	while (true)
	{
		byte tempCharacter = *text;
		text += 1;
		if (tempCharacter == 0)
		{
			return;
		}
		*videoBufferPointer = tempCharacter;
		videoBufferPointer += 1;
	}
}

void drawTextFromProgMem(byte posX, byte posY, const byte *text)
{
	byte tempBuffer[200];
	readTextFromProgMem(tempBuffer, text);
	drawText(posX, posY, tempBuffer);
}

short getNumberOfFiles()
{
    return numberOfFiles;
}

// Returns the address of the file entry at the given index.
long getFileEntryAddressByIndex(short index)
{
    if (index >= getNumberOfFiles())
    {
        return -1;
    }
    short tempCount = 0;
    long output = 0;
    while (true)
    {
        byte tempCharacter = fileNameList[output];
        if (tempCharacter != 0)
        {
            if (tempCount >= index)
            {
                return output;
            }
            tempCount += 1;
        }
        output += MAXIMUM_FILE_NAME_LENGTH;
    }
}

long getNextFileEntryAddress(long fileEntryAddress)
{
    if (fileEntryAddress < 0)
    {
        return -1;
    }
    fileEntryAddress += MAXIMUM_FILE_NAME_LENGTH;
    while (true)
    {
        byte tempCharacter = fileNameList[fileEntryAddress];
        if (tempCharacter != 0)
        {
            return fileEntryAddress;
        }
        fileEntryAddress += MAXIMUM_FILE_NAME_LENGTH;
        if (fileEntryAddress >= sizeof(fileNameList))
        {
            return -1;
        }
    }
}

void getFileName(byte *destination, long fileEntryAddress)
{
    if (fileEntryAddress < 0)
    {
        *destination = 0;
    } else {
        copyText(destination, fileNameList + fileEntryAddress);
    }
}

long getFileEntryAddressByName(byte *name)
{
    short index = 0;
    long tempAddress = getFileEntryAddressByIndex(index);
    while (true)
    {
        byte tempName[50];
        getFileName(tempName, tempAddress);
        if (equalText(tempName, name))
        {
            return tempAddress;
        }
        index += 1;
        tempAddress = getNextFileEntryAddress(tempAddress);
        if (tempAddress == -1)
        {
            return -1;
        }
    }
}

void printFileNameListToConsole()
{
    short index = 0;
    long tempAddress = getFileEntryAddressByIndex(index);
    while (index < numberOfFiles)
    {
        byte tempName[50];
        getFileName(tempName, tempAddress);
        cout << (char *)tempName << endl;
        index += 1;
        tempAddress = getNextFileEntryAddress(tempAddress);
    }
    cout << endl;
}

void getFilePathByName(byte *destination, byte *name)
{
    copyText(destination, (byte *)SD_CARD_DIRECTORY_PATH);
    short tempIndex = sizeof(SD_CARD_DIRECTORY_PATH);
    *(destination + tempIndex - 1) = '/';
    copyText(destination + tempIndex, name);
}

void getFilePathByEntryAddress(byte *destination, long fileEntryAddress)
{
    byte tempName[50];
    getFileName(tempName, fileEntryAddress);
    getFilePathByName(destination, tempName);
}

long getFileSize(long fileEntryAddress)
{
    if (fileEntryAddress < 0)
    {
        return -1;
    }
    byte tempPath[500];
    getFilePathByEntryAddress(tempPath, fileEntryAddress);
    ifstream tempFile((char *)tempPath, ifstream::in | ifstream::binary | ifstream::ate);
    return (long)(tempFile.tellg());
}

void readFile(unsigned short sramDestinationAddress, long fileEntryAddress)
{
    if (fileEntryAddress < 0)
    {
        return;
    }
    long tempSize = getFileSize(fileEntryAddress);
    FILE * tempFile;
    byte tempPath[500];
    getFilePathByEntryAddress(tempPath, fileEntryAddress);
    tempFile = fopen((char *)tempPath, "rb");
    fread((void *)(externalSram + sramDestinationAddress), 1, tempSize, tempFile);
    fclose(tempFile);
}

void writeFile(long fileEntryAddress, unsigned short sramSourceAddress)
{
    if (fileEntryAddress < 0)
    {
        return;
    }
    long tempSize = getFileSize(fileEntryAddress);
    FILE * tempFile;
    byte tempPath[500];
    getFilePathByEntryAddress(tempPath, fileEntryAddress);
    tempFile = fopen((char *)tempPath, "wb");
    fwrite((void *)(externalSram + sramSourceAddress), 1, tempSize, tempFile);
    fclose(tempFile);
}

long createFile(byte *name, unsigned long size)
{
    FILE * tempFile;
    byte tempPath[500];
    getFilePathByName(tempPath, name);
    tempFile = fopen((char *)tempPath, "wb");
    byte tempData = 0;
    unsigned long tempCount = 0;
    while (tempCount < size)
    {
        fwrite((void *)(&tempData), 1, 1, tempFile);
        tempCount += 1;
    }
    fclose(tempFile);
    long output = 0;
    while (true)
    {
        byte tempCharacter = fileNameList[output];
        if (tempCharacter == 0)
        {
            break;
        }
        output += MAXIMUM_FILE_NAME_LENGTH;
    }
    copyText(fileNameList + output, name);
    numberOfFiles += 1;
    return output;
}

void deleteFile(long fileEntryAddress)
{
    if (fileEntryAddress < 0)
    {
        return;
    }
    byte tempPath[500];
    getFilePathByEntryAddress(tempPath, fileEntryAddress);
    remove((char *)tempPath);
    *(fileNameList + fileEntryAddress) = 0;
    numberOfFiles -= 1;
}

void startReadingFromSram(long address)
{
    if (externalSramState == 0)
    {
        externalSramState = 1;
    } else {
        cout << "WARNING: Trying to read from SRAM when not in closed state!" << endl;
    }
    externalSramAddress = address;
}

void startWritingToSram(long address)
{
    if (externalSramState == 0)
    {
        externalSramState = 2;
    } else {
        cout << "WARNING: Trying to write to SRAM when not in closed state!" << endl;
    }
    externalSramAddress = address;
}

byte receiveByteFromSram()
{
    byte output = 0;
    if (externalSramAddress < EXTERNAL_SRAM_SIZE)
    {
        output = *(externalSram + externalSramAddress);
    }
    if (externalSramAddress == PROGRAM_COUNTER_ADDRESS)
    {
        output = (programCounter & 0x00FF);
    }
    if (externalSramAddress == PROGRAM_COUNTER_ADDRESS + 1)
    {
        output = (programCounter & 0xFF00) >> 8;
    }
    if (externalSramAddress == GLOBAL_SCOPE_ADDRESS_ADDRESS)
    {
        output = (globalScopeAddress & 0x00FF);
    }
    if (externalSramAddress == GLOBAL_SCOPE_ADDRESS_ADDRESS + 1)
    {
        output = (globalScopeAddress & 0xFF00) >> 8;
    }
    if (externalSramAddress == LOCAL_SCOPE_ADDRESS_ADDRESS)
    {
        output = (localScopeAddress & 0x00FF);
    }
    if (externalSramAddress == LOCAL_SCOPE_ADDRESS_ADDRESS + 1)
    {
        output = (localScopeAddress & 0xFF00) >> 8;
    }
    externalSramAddress += 1;
    return output;
}

unsigned short receiveShortFromSram()
{
    unsigned short output = 0;
    output |= receiveByteFromSram();
    output |= receiveByteFromSram() << 8;
    return output;
}

unsigned long receiveLongFromSram()
{
    unsigned long output = 0;
    output |= (long)receiveByteFromSram();
    output |= (long)receiveByteFromSram() << (long)8;
    output |= (long)receiveByteFromSram() << (long)16;
    output |= (long)receiveByteFromSram() << (long)24;
    return output;
}

unsigned long receiveNumberFromSram(unsigned short size)
{
    if (size == 1)
    {
        return receiveByteFromSram();
    }
    if (size == 2)
    {
        return receiveShortFromSram();
    }
    if (size == 4)
    {
        return receiveLongFromSram();
    }
    return -1;
}

long receiveSignedNumberFromSram(unsigned short size)
{
    if (size == 1)
    {
        return (char)receiveByteFromSram();
    }
    if (size == 2)
    {
        return (short)receiveShortFromSram();
    }
    if (size == 4)
    {
        return (long)receiveLongFromSram();
    }
    return -1;
}

void receiveDataFromSram(byte *destination, long amount)
{
    while (amount > 0)
    {
        *destination = receiveByteFromSram();
        destination += 1;
        amount -= 1;
    }
}

void sendByteToSram(byte data)
{
    if (externalSramAddress < EXTERNAL_SRAM_SIZE)
    {
        *(externalSram + externalSramAddress) = data;
    }
    if (externalSramAddress == PROGRAM_COUNTER_ADDRESS)
    {
        programCounter &= 0xFF00;
        programCounter |= data;
    }
    if (externalSramAddress == PROGRAM_COUNTER_ADDRESS + 1)
    {
        programCounter &= 0x00FF;
        programCounter |= data << 8;
    }
    if (externalSramAddress == GLOBAL_SCOPE_ADDRESS_ADDRESS)
    {
        globalScopeAddress &= 0xFF00;
        globalScopeAddress |= data;
    }
    if (externalSramAddress == GLOBAL_SCOPE_ADDRESS_ADDRESS + 1)
    {
        globalScopeAddress &= 0x00FF;
        globalScopeAddress |= data << 8;
    }
    if (externalSramAddress == LOCAL_SCOPE_ADDRESS_ADDRESS)
    {
        localScopeAddress &= 0xFF00;
        localScopeAddress |= data;
    }
    if (externalSramAddress == LOCAL_SCOPE_ADDRESS_ADDRESS + 1)
    {
        localScopeAddress &= 0x00FF;
        localScopeAddress |= data << 8;
    }
    externalSramAddress += 1;
}

void sendShortToSram(unsigned short data)
{
    sendByteToSram(data & 0x00FF);
    sendByteToSram((data & 0xFF00) >> 8);
}

void sendLongToSram(unsigned long data)
{
    sendByteToSram(data & 0x000000FF);
    sendByteToSram((data & 0x0000FF00) >> 8);
    sendByteToSram((data & 0x00FF0000) >> 16);
    sendByteToSram((data & 0xFF000000) >> 24);
}

void sendNumberToSram(unsigned short size, unsigned long number)
{
    if (size == 1)
    {
        sendByteToSram(number);
    }
    if (size == 2)
    {
        sendShortToSram(number);
    }
    if (size == 4)
    {
        sendLongToSram(number);
    }
}

void sendDataToSram(long amount, byte *source)
{
    while (amount > 0)
    {
        sendByteToSram(*source);
        source += 1;
        amount -= 1;
    }
}

void stopConnectionToSram()
{
    externalSramState = 0;
    // Pull CS pin high.
}

unsigned long readNumberFromSram(unsigned short size, long address)
{
    startReadingFromSram(address);
    unsigned long output = receiveNumberFromSram(size);
    stopConnectionToSram();
    return output;
}

long readSignedNumberFromSram(unsigned short size, long address)
{
    startReadingFromSram(address);
    long output = receiveSignedNumberFromSram(size);
    stopConnectionToSram();
    return output;
}

float readFloatFromSram(long address)
{
    unsigned long tempNumber = readNumberFromSram(4, address);
    return *(float *)&(tempNumber);
}

void readTextFromSram(byte *destination, long address)
{
    startReadingFromSram(address);
    unsigned short index = 0;
    while (true)
    {
        byte tempCharacter = receiveByteFromSram();
        *(destination + index) = tempCharacter;
        if (tempCharacter == 0)
        {
            break;
        }
        index += 1;
    }
    stopConnectionToSram();
}

void writeNumberToSram(long address, unsigned short size, unsigned long number)
{
    startWritingToSram(address);
    sendNumberToSram(size, number);
    stopConnectionToSram();
}

void writeFloatToSram(long address, float number)
{
    unsigned long tempNumber = *(unsigned long *)&(number);
    writeNumberToSram(address, 4, tempNumber);
}

void writeTextToSram(long address, byte *text)
{
    startWritingToSram(address);
    unsigned short index = 0;
    while (true)
    {
        byte tempCharacter = *(text + index);
        sendByteToSram(tempCharacter);
        if (tempCharacter == 0)
        {
            break;
        }
        index += 1;
    }
    stopConnectionToSram();
}

void copyFromSramToSram(long destinationAddress, unsigned short amount, long sourceAddress)
{
    if (destinationAddress < sourceAddress)
    {
        unsigned short index = 0;
        while (index < amount)
        {
            byte tempBuffer[SRAM_BUFFER_SIZE];
            unsigned short tempLength;
            if (amount < SRAM_BUFFER_SIZE || index > amount - SRAM_BUFFER_SIZE)
            {
                tempLength = amount - index;
            } else {
                tempLength = SRAM_BUFFER_SIZE;
            }
            startReadingFromSram(sourceAddress + index);
            receiveDataFromSram(tempBuffer, tempLength);
            stopConnectionToSram();
            startWritingToSram(destinationAddress + index);
            sendDataToSram(tempLength, tempBuffer);
            stopConnectionToSram();
            index += tempLength;
        }
    } else {
        unsigned short index = amount;
        while (index > 0)
        {
            byte tempBuffer[SRAM_BUFFER_SIZE];
            unsigned short tempLength;
            if (index < SRAM_BUFFER_SIZE)
            {
                tempLength = index;
            } else {
                tempLength = SRAM_BUFFER_SIZE;
            }
            index -= tempLength;
            startReadingFromSram(sourceAddress + index);
            receiveDataFromSram(tempBuffer, tempLength);
            stopConnectionToSram();
            startWritingToSram(destinationAddress + index);
            sendDataToSram(tempLength, tempBuffer);
            stopConnectionToSram();
        }
    }
}

void runFile(long fileEntryAddress)
{
    srand(getTimeMilliseconds());
    // Necessary to begin the random number generation.
    long tempTrash = rand();
    long tempSize = getFileSize(fileEntryAddress);
    readFile(0, fileEntryAddress);
    startReadingFromSram(0);
    addressSize = receiveByteFromSram();
    short globalScopeSize = receiveNumberFromSram(addressSize);
    stopConnectionToSram();
    programCounter = 1 + addressSize;
    globalScopeAddress = tempSize;
    localScopeAddress = globalScopeAddress;
    startWritingToSram(localScopeAddress);
    sendByteToSram(addressSize);
    sendNumberToSram(addressSize, 0);
    sendNumberToSram(addressSize, localScopeAddress);
    sendNumberToSram(addressSize, globalScopeSize);
    stopConnectionToSram();
    startWritingToSram(HEAP_START_ADDRESS - 6);
    sendShortToSram(0);
    sendShortToSram(0);
    sendShortToSram(0);
    stopConnectionToSram();
    runtimeMode = 0;
    clearVideoBuffer();
}

byte convertNumberToHexDigit(byte number)
{
    if (number < 10)
    {
        return number + '0';
    } else {
        return (number - 10) + 'A';
    }
}

byte convertHexDigitToNumber(byte digit)
{
    if (digit >= '0' && digit <= '9')
    {
        return digit - '0';
    }
    if (digit >= 'A' && digit <= 'F')
    {
        return (digit - 'A') + 10;
    }
    if (digit >= 'a' && digit <= 'f')
    {
        return (digit - 'a') + 10;
    }
    return 0;
}

// Direction values:
// 0 = left
// 1 = right
// 2 = up
// 3 = down
void scrollVideoBuffer(byte direction, byte *text)
{
    switch (direction)
    {
        case 0:
        {
            byte tempPosY = 0;
            while (tempPosY < VIDEO_CHARACTER_HEIGHT)
            {
                byte tempBuffer[VIDEO_CHARACTER_WIDTH - 1];
                copyData(tempBuffer, VIDEO_CHARACTER_WIDTH - 1, videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH + 1);
                copyData(videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH, VIDEO_CHARACTER_WIDTH - 1, tempBuffer);
                *(videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH + VIDEO_CHARACTER_WIDTH - 1) = *(text + tempPosY);
                tempPosY += 1;
            }
            break;
        }
        case 1:
        {
            byte tempPosY = 0;
            while (tempPosY < VIDEO_CHARACTER_HEIGHT)
            {
                byte tempBuffer[VIDEO_CHARACTER_WIDTH - 1];
                copyData(tempBuffer, VIDEO_CHARACTER_WIDTH - 1, videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH);
                copyData(videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH + 1, VIDEO_CHARACTER_WIDTH - 1, tempBuffer);
                *(videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH) = *(text + tempPosY);
                tempPosY += 1;
            }
            break;
        }
        case 2:
        {
            byte tempPosY = 1;
            while (tempPosY < VIDEO_CHARACTER_HEIGHT)
            {
                copyData(videoBuffer + (short)(tempPosY - 1) * VIDEO_CHARACTER_WIDTH, VIDEO_CHARACTER_WIDTH, videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH);
                tempPosY += 1;
            }
            copyData(videoBuffer + (VIDEO_CHARACTER_HEIGHT - 1) * VIDEO_CHARACTER_WIDTH, VIDEO_CHARACTER_WIDTH, text);
            break;
        }
        case 3:
        {
            byte tempPosY = VIDEO_CHARACTER_HEIGHT;
            while (tempPosY > 0)
            {
                copyData(videoBuffer + (short)tempPosY * VIDEO_CHARACTER_WIDTH, VIDEO_CHARACTER_WIDTH, videoBuffer + (short)(tempPosY - 1) * VIDEO_CHARACTER_WIDTH);
                tempPosY -= 1;
            }
            copyData(videoBuffer, VIDEO_CHARACTER_WIDTH, text);
            break;
        }
        default:
        {
            
            break;
        }
    }
}

// Returns either 0 or 1.
byte gameOfLifeCellIsAlive(short posX, short posY, short width, short height, unsigned short gridAddress)
{
    if (posX < 0 || posX >= width || posY < 0 || posY >= height)
    {
        return 0;
    }
    unsigned short index = posX + posY * width;
    byte tempCharacter = readNumberFromSram(1, gridAddress + index);
    if (tempCharacter == GAME_OF_LIFE_DEAD_CELL)
    {
        return 0;
    } else {
        return 1;
    }
}

// Returns either 0 or 1.
byte computeMandelbrotPoint(float realPart, float imaginaryPart, unsigned short count)
{
    float tempLastRealPart = realPart;
    float tempLastImaginaryPart = imaginaryPart;
    byte output = 1;
    while (count > 0)
    {
        float tempNextRealPart = tempLastRealPart * tempLastRealPart - tempLastImaginaryPart * tempLastImaginaryPart + realPart;
        float tempNextImaginaryPart = 2 * tempLastRealPart * tempLastImaginaryPart + imaginaryPart;
        if (tempNextRealPart > 2 || tempNextRealPart < -2 || tempNextImaginaryPart > 2 || tempNextImaginaryPart < -2)
        {
            output = 0;
            break;
        }
        tempLastRealPart = tempNextRealPart;
        tempLastImaginaryPart = tempNextImaginaryPart;
        count -= 1;
    }
    return output;
}

void quitCurrentProgram()
{
    byte tempName[50];
    readTextFromProgMem(tempName, START_FILE_NAME);
    long tempAddress = getFileEntryAddressByName(tempName);
    runFile(tempAddress);
}

void displayPromptSelection()
{
    clearVideoBuffer();
    promptIndex2 = promptIndex - promptIndex % VIDEO_CHARACTER_HEIGHT;
    startReadingFromSram(promptSourceAddress);
    unsigned short index = 0;
    while (index < promptIndex2)
    {
        byte tempCharacter = receiveByteFromSram();
        if (tempCharacter == 0)
        {
            index += 1;
        }
    }
    unsigned short tempVideoBufferIndex = 1;
    unsigned short offset = 0;
    while (offset < VIDEO_CHARACTER_HEIGHT && promptIndex2 + offset < promptAmount)
    {
        byte tempCharacter = receiveByteFromSram();
        if (tempCharacter == 0)
        {
            offset += 1;
            tempVideoBufferIndex = 1 + offset * VIDEO_CHARACTER_WIDTH;
        } else {
            videoBuffer[tempVideoBufferIndex] = tempCharacter;
            tempVideoBufferIndex += 1;
        }
    }
    stopConnectionToSram();
    offset = promptIndex - promptIndex2;
    videoBuffer[offset * VIDEO_CHARACTER_WIDTH] = PROMPT_SELECTION_CURSOR_CHARACTER;
}

void displayPromptFile()
{
    clearVideoBuffer();
    promptIndex2 = promptIndex - promptIndex % VIDEO_CHARACTER_HEIGHT;
    byte tempBuffer[50];
    long tempAddress = getFileEntryAddressByIndex(promptIndex2);
    unsigned short tempVideoBufferIndex = 1;
    unsigned short offset = 0;
    while (offset < VIDEO_CHARACTER_HEIGHT && promptIndex2 + offset < promptAmount)
    {
        getFileName(tempBuffer, tempAddress);
        tempBuffer[VIDEO_CHARACTER_WIDTH - 1] = 0;
        tempVideoBufferIndex = 1 + offset * VIDEO_CHARACTER_WIDTH;
        unsigned short index = 0;
        while (true)
        {
            byte tempCharacter = tempBuffer[index];
            if (tempCharacter == 0)
            {
                break;
            }
            videoBuffer[tempVideoBufferIndex] = tempCharacter;
            tempVideoBufferIndex += 1;
            index += 1;
        }
        tempAddress = getNextFileEntryAddress(tempAddress);
        offset += 1;
    }
    offset = promptIndex - promptIndex2;
    videoBuffer[offset * VIDEO_CHARACTER_WIDTH] = PROMPT_SELECTION_CURSOR_CHARACTER;
}

// Upon invocation, SRAM should be reading sequentially
// starting at the program counter.
long getNextArgumentAddress(byte shouldStopConnectionToSram)
{
    byte tempArgumentType = receiveByteFromSram();
    programCounter += 1;
    byte tempTypeIdentifier = (tempArgumentType & 0xF0);
    byte tempShouldLoadConstantAsArgument = (tempArgumentType & 0x08);
    unsigned short tempArgumentSize = (tempArgumentType & 0x07);
    byte tempSramAddressIsEqualToProgramCounter;
    if (tempArgumentSize == 0 && tempTypeIdentifier < 0x40)
    {
        unsigned short tempAddress = getNextArgumentAddress(true);
        stopConnectionToSram();
        startReadingFromSram(tempAddress);
        tempArgumentSize = receiveNumberFromSram(addressSize);
        tempSramAddressIsEqualToProgramCounter = false;
    } else {
        tempSramAddressIsEqualToProgramCounter = true;
    }
    long tempConstantAddress;
    if (tempShouldLoadConstantAsArgument)
    {
        if (!tempSramAddressIsEqualToProgramCounter)
        {
            stopConnectionToSram();
            startReadingFromSram(programCounter);
            tempSramAddressIsEqualToProgramCounter = true;
        }
        tempConstantAddress = getNextArgumentAddress(false);
    } else {
        tempConstantAddress = programCounter;
    }
    long output = -1;
    if (tempTypeIdentifier < 0x10 || tempTypeIdentifier > 0x30)
    {
        if (tempTypeIdentifier == 0x00)
        {
            output = tempConstantAddress;
            programCounter += tempArgumentSize;
            tempSramAddressIsEqualToProgramCounter = false;
        }
        // These argument types should never accept a constant.
        if (tempTypeIdentifier == 0x40)
        {
            output = PROGRAM_COUNTER_ADDRESS;
        }
        if (tempTypeIdentifier == 0x50)
        {
            output = GLOBAL_SCOPE_ADDRESS_ADDRESS;
        }
        if (tempTypeIdentifier == 0x60)
        {
            output = LOCAL_SCOPE_ADDRESS_ADDRESS;
        }
    } else {
        if (tempShouldLoadConstantAsArgument)
        {
            stopConnectionToSram();
            startReadingFromSram(tempConstantAddress);
            tempSramAddressIsEqualToProgramCounter = false;
        } else if (!tempSramAddressIsEqualToProgramCounter)
        {
            stopConnectionToSram();
            startReadingFromSram(programCounter);
            tempSramAddressIsEqualToProgramCounter = true;
        }
        if (tempTypeIdentifier == 0x10)
        {
            output = receiveNumberFromSram(tempArgumentSize);
        }
        if (tempTypeIdentifier == 0x20)
        {
            output = globalScopeAddress + receiveNumberFromSram(tempArgumentSize);
        }
        if (tempTypeIdentifier == 0x30)
        {
            output = localScopeAddress + receiveNumberFromSram(tempArgumentSize);
        }
        if (!tempShouldLoadConstantAsArgument)
        {
            programCounter += tempArgumentSize;
        }
    }
    if (shouldStopConnectionToSram)
    {
        stopConnectionToSram();
    } else {
        if (!tempSramAddressIsEqualToProgramCounter)
        {
            stopConnectionToSram();
            startReadingFromSram(programCounter);
            tempSramAddressIsEqualToProgramCounter = true;
        }
    }
    return output;
}

void processNextCommand()
{
    startReadingFromSram(programCounter);
    byte tempOpcode = receiveByteFromSram();
    programCounter += 1;
    byte tempFunctionAddressSize = 2;
    byte numberOfArguments = pgm_read_byte(COMMAND_ARGUMENT_AMOUNT_LIST + tempOpcode);
    long argumentAddressList[25];
    byte index = 0;
    while (index < numberOfArguments)
    {
        byte shouldStopConnectionToSram;
        if (index < numberOfArguments - 1)
        {
            shouldStopConnectionToSram = false;
        } else {
            shouldStopConnectionToSram = true;
        }
        long tempAddress = getNextArgumentAddress(shouldStopConnectionToSram);
        argumentAddressList[index] = tempAddress;
        // FNC.
        if (tempOpcode == 19 && index == 0)
        {
            stopConnectionToSram();
            tempFunctionAddressSize = readNumberFromSram(1, tempAddress);
            numberOfArguments += readNumberFromSram(tempFunctionAddressSize, tempAddress + 1 + 2 * tempFunctionAddressSize);
            startReadingFromSram(programCounter);
        }
        index += 1;
    }
    stopConnectionToSram();
    switch (tempOpcode)
    {
        // WRT.
        case 0:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempSourceAddress = argumentAddressList[1];
            long tempDestinationAddress = argumentAddressList[2];
            copyFromSramToSram(tempDestinationAddress, tempLength, tempSourceAddress);
            break;
        }
        // NOT.
        case 1:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = ~tempData;
            writeNumberToSram(argumentAddressList[2], tempLength, tempResult);
            break;
        }
        // WRZ.
        case 2:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempCondition = readNumberFromSram(tempLength1, argumentAddressList[1]);
            if (!tempCondition)
            {
                unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
                unsigned long tempData = readNumberFromSram(tempLength2, argumentAddressList[3]);
                writeNumberToSram(argumentAddressList[4], tempLength2, tempData);
            }
            break;
        }
        // WNZ.
        case 3:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempCondition = readNumberFromSram(tempLength1, argumentAddressList[1]);
            if (tempCondition)
            {
                unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
                unsigned long tempData = readNumberFromSram(tempLength2, argumentAddressList[3]);
                writeNumberToSram(argumentAddressList[4], tempLength2, tempData);
            }
            break;
        }
        // OR.
        case 4:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempData2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = (tempData1 | tempData2);
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // AND.
        case 5:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempData2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = (tempData1 & tempData2);
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // XOR.
        case 6:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempData2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = (tempData1 ^ tempData2);
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // BNT.
        case 7:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            byte tempResult = !tempNumber;
            writeNumberToSram(argumentAddressList[2], 1, tempResult);
            break;
        }
        // GRE.
        case 8:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            byte tempResult = (tempNumber1 > tempNumber2);
            writeNumberToSram(argumentAddressList[3], 1, tempResult);
            break;
        }
        // GRES.
        case 9:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber1 = readSignedNumberFromSram(tempLength, argumentAddressList[1]);
            long tempNumber2 = readSignedNumberFromSram(tempLength, argumentAddressList[2]);
            byte tempResult = (tempNumber1 > tempNumber2);
            writeNumberToSram(argumentAddressList[3], 1, tempResult);
            break;
        }
        // EQU.
        case 10:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            byte tempResult = (tempNumber1 == tempNumber2);
            writeNumberToSram(argumentAddressList[3], 1, tempResult);
            break;
        }
        // ADD.
        case 11:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = tempNumber1 + tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // SUB.
        case 12:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = tempNumber1 - tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // BSL.
        case 13:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            byte tempNumber2 = readNumberFromSram(1, argumentAddressList[2]);
            unsigned long tempResult = (tempNumber1 << tempNumber2);
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // BSR.
        case 14:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            byte tempNumber2 = readNumberFromSram(1, argumentAddressList[2]);
            unsigned long tempResult = (tempNumber1 >> tempNumber2);
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // INC.
        case 15:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = tempNumber1 + 1;
            writeNumberToSram(argumentAddressList[1], tempLength, tempResult);
            break;
        }
        // DEC.
        case 16:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = tempNumber1 - 1;
            writeNumberToSram(argumentAddressList[1], tempLength, tempResult);
            break;
        }
        // SLO.
        case 17:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = (tempNumber1 << 1);
            writeNumberToSram(argumentAddressList[1], tempLength, tempResult);
            break;
        }
        // SRO.
        case 18:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = (tempNumber1 >> 1);
            writeNumberToSram(argumentAddressList[1], tempLength, tempResult);
            break;
        }
        // FNC.
        case 19:
        {
            long tempFunctionInfoAddress = argumentAddressList[0];
            startReadingFromSram(tempFunctionInfoAddress + 1);
            unsigned short tempCodeAddress = receiveNumberFromSram(tempFunctionAddressSize);
            unsigned short tempScopeSize = receiveNumberFromSram(tempFunctionAddressSize);
            stopConnectionToSram();
            unsigned short tempScopeSize2 = readNumberFromSram(addressSize, localScopeAddress + 1 + 2 * addressSize);
            unsigned short tempScopeAddress = localScopeAddress + tempScopeSize2;
            startWritingToSram(tempScopeAddress);
            sendByteToSram(tempFunctionAddressSize);
            sendNumberToSram(tempFunctionAddressSize, programCounter);
            sendNumberToSram(tempFunctionAddressSize, localScopeAddress);
            sendNumberToSram(tempFunctionAddressSize, tempScopeSize);
            byte index = 1;
            while (index < numberOfArguments)
            {
                sendNumberToSram(tempFunctionAddressSize, argumentAddressList[index]);
                index += 1;
            }
            stopConnectionToSram();
            programCounter = tempCodeAddress;
            localScopeAddress = tempScopeAddress;
            addressSize = tempFunctionAddressSize;
            break;
        }
        // RET.
        case 20:
        {
            startReadingFromSram(localScopeAddress + 1);
            programCounter = receiveNumberFromSram(addressSize);
            localScopeAddress = receiveNumberFromSram(addressSize);
            stopConnectionToSram();
            startReadingFromSram(localScopeAddress);
            addressSize = receiveByteFromSram();
            stopConnectionToSram();
            break;
        }
        // FIL.
        case 21:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData = readNumberFromSram(tempLength1, argumentAddressList[1]);
            unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
            unsigned short tempAmount = readNumberFromSram(tempLength2, argumentAddressList[3]);
            startWritingToSram(argumentAddressList[4]);
            unsigned short index = 0;
            while (index < tempAmount)
            {
                sendNumberToSram(tempLength1, tempData);
                index += 1;
            }
            stopConnectionToSram();
            break;
        }
        // FND.
        case 22:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempData1 = readNumberFromSram(tempLength1, argumentAddressList[1]);
            unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
            unsigned long tempSize = readNumberFromSram(tempLength2, argumentAddressList[3]);
            byte hasFoundPattern = false;
            unsigned long index = 0;
            while (index < tempSize - tempLength1 + 1)
            {
                unsigned long tempData2 = readNumberFromSram(tempLength1, argumentAddressList[4] + index);
                if (tempData2 == tempData1)
                {
                    hasFoundPattern = true;
                    break;
                }
                index += 1;
            }
            if (!hasFoundPattern)
            {
                index = 0xFFFFFFFF;
            }
            writeNumberToSram(argumentAddressList[5], tempLength2, index);
            break;
        }
        // CII.
        case 23:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength1, argumentAddressList[1]);
            unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
            writeNumberToSram(argumentAddressList[3], tempLength2, tempNumber);
            break;
        }
        // CIIS.
        case 24:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber = readSignedNumberFromSram(tempLength1, argumentAddressList[1]);
            unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[2]);
            writeNumberToSram(argumentAddressList[3], tempLength2, tempNumber);
            break;
        }
        // MUL.
        case 25:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = tempNumber1 * tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // MULS.
        case 26:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber1 = readSignedNumberFromSram(tempLength, argumentAddressList[1]);
            long tempNumber2 = readSignedNumberFromSram(tempLength, argumentAddressList[2]);
            long tempResult = tempNumber1 * tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // DIV.
        case 27:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = tempNumber1 / tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // DIVS.
        case 28:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber1 = readSignedNumberFromSram(tempLength, argumentAddressList[1]);
            long tempNumber2 = readSignedNumberFromSram(tempLength, argumentAddressList[2]);
            long tempResult = tempNumber1 / tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // MOD.
        case 29:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber1 = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempNumber2 = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned long tempResult = tempNumber1 % tempNumber2;
            writeNumberToSram(argumentAddressList[3], tempLength, tempResult);
            break;
        }
        // ADDF.
        case 30:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = tempNumber1 + tempNumber2;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // SUBF.
        case 31:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = tempNumber1 - tempNumber2;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // MULF.
        case 32:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = tempNumber1 * tempNumber2;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // DIVF.
        case 33:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = tempNumber1 / tempNumber2;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // CIF.
        case 34:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            float tempResult = (float)tempNumber;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // CIFS.
        case 35:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber = readSignedNumberFromSram(tempLength, argumentAddressList[1]);
            float tempResult = (float)tempNumber;
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // CFI.
        case 36:
        {
            float tempNumber = readFloatFromSram(argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned long tempResult = (unsigned long)tempNumber;
            writeNumberToSram(argumentAddressList[2], tempLength, tempResult);
            break;
        }
        // CFIS.
        case 37:
        {
            float tempNumber = readFloatFromSram(argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            long tempResult = (long)tempNumber;
            writeNumberToSram(argumentAddressList[2], tempLength, tempResult);
            break;
        }
        // CIT.
        case 38:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            byte tempBuffer[20];
            convertLongToText(tempBuffer, tempNumber);
            writeTextToSram(argumentAddressList[2], tempBuffer);
            break;
        }
        // CITS.
        case 39:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempNumber = readSignedNumberFromSram(tempLength, argumentAddressList[1]);
            byte tempBuffer[20];
            convertLongToText(tempBuffer, tempNumber);
            writeTextToSram(argumentAddressList[2], tempBuffer);
            break;
        }
        // CFT.
        case 40:
        {
            float tempNumber = readFloatFromSram(argumentAddressList[0]);
            byte tempBuffer[20];
            convertFloatToText(tempBuffer, tempNumber);
            writeTextToSram(argumentAddressList[1], tempBuffer);
            break;
        }
        // CTI.
        case 41:
        {
            byte tempBuffer[20];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned long tempNumber = convertTextToLong(tempBuffer);
            writeNumberToSram(argumentAddressList[2], tempLength, tempNumber);
            break;
        }
        // CTIS.
        case 42:
        {
            byte tempBuffer[20];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            long tempNumber = convertTextToLong(tempBuffer);
            writeNumberToSram(argumentAddressList[2], tempLength, tempNumber);
            break;
        }
        // CTF.
        case 43:
        {
            byte tempBuffer[20];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            float tempNumber = convertTextToFloat(tempBuffer);
            writeFloatToSram(argumentAddressList[1], tempNumber);
            break;
        }
        // WRTT.
        case 44:
        {
            unsigned short index = 0;
            while (true)
            {
                byte tempCharacter = readNumberFromSram(1, argumentAddressList[0] + index);
                writeNumberToSram(argumentAddressList[1] + index, 1, tempCharacter);
                if (tempCharacter == 0)
                {
                    break;
                }
                index += 1;
            }
            break;
        }
        // EQUT.
        case 45:
        {
            byte tempResult = true;
            unsigned short index = 0;
            while (true)
            {
                byte tempCharacter1 = readNumberFromSram(1, argumentAddressList[0] + index);
                byte tempCharacter2 = readNumberFromSram(1, argumentAddressList[1] + index);
                if (tempCharacter1 != tempCharacter2)
                {
                    tempResult = false;
                    break;
                }
                if (tempCharacter1 == 0)
                {
                    break;
                }
                index += 1;
            }
            writeNumberToSram(argumentAddressList[2], 1, tempResult);
            break;
        }
        // FNDT.
        case 46:
        {
            byte hasFoundPattern = false;
            byte hasReachedEndOfText = false;
            unsigned short index = 0;
            while (!hasReachedEndOfText)
            {
                unsigned short tempOffset = 0;
                while (true)
                {
                    byte tempCharacter1 = readNumberFromSram(1, argumentAddressList[0] + tempOffset);
                    byte tempCharacter2 = readNumberFromSram(1, argumentAddressList[1] + index + tempOffset);
                    if (tempCharacter1 == 0)
                    {
                        hasFoundPattern = true;
                        break;
                    }
                    if (tempCharacter2 == 0)
                    {
                        hasReachedEndOfText = true;
                        break;
                    }
                    if (tempCharacter1 != tempCharacter2)
                    {
                        break;
                    }
                    tempOffset += 1;
                }
                if (hasFoundPattern || hasReachedEndOfText)
                {
                    break;
                }
                index += 1;
            }
            if (!hasFoundPattern)
            {
                index = 0xFFFF;
            }
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[2]);
            writeNumberToSram(argumentAddressList[3], tempLength, index);
            break;
        }
        // LENT.
        case 47:
        {
            startReadingFromSram(argumentAddressList[0]);
            unsigned short index = 0;
            while (true)
            {
                byte tempCharacter = receiveByteFromSram();
                if (tempCharacter == 0)
                {
                    break;
                }
                index += 1;
            }
            stopConnectionToSram();
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            writeNumberToSram(argumentAddressList[2], tempLength, index);
            break;
        }
        // SUBT.
        case 48:
        {
            unsigned short tempOffset = 0;
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned short index = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned short tempEndIndex = readNumberFromSram(tempLength, argumentAddressList[3]);
            while (index < tempEndIndex)
            {
                byte tempCharacter = readNumberFromSram(1, argumentAddressList[0] + index);
                writeNumberToSram(argumentAddressList[4] + tempOffset, 1, tempCharacter);
                tempOffset += 1;
                index += 1;
            }
            writeNumberToSram(argumentAddressList[4] + tempOffset, 1, 0);
            break;
        }
        // CONT.
        case 49:
        {
            startReadingFromSram(argumentAddressList[1]);
            unsigned short index = 0;
            while (true)
            {
                byte tempCharacter = receiveByteFromSram();
                if (tempCharacter == 0)
                {
                    break;
                }
                index += 1;
            }
            stopConnectionToSram();
            unsigned short tempOffset = 0;
            while (true)
            {
                byte tempCharacter = readNumberFromSram(1, argumentAddressList[0] + tempOffset);
                writeNumberToSram(argumentAddressList[1] + index + tempOffset, 1, tempCharacter);
                if (tempCharacter == 0)
                {
                    break;
                }
                tempOffset += 1;
            }
            break;
        }
        // SDL.
        case 50:
        {
            unsigned short tempResult = 1;
            startReadingFromSram(argumentAddressList[0]);
            while (true)
            {
                byte tempCharacter = receiveByteFromSram();
                if (tempCharacter == ' ')
                {
                    tempResult += 1;
                }
                if (tempCharacter == 0)
                {
                    break;
                }
            }
            stopConnectionToSram();
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            writeNumberToSram(argumentAddressList[2], tempLength, tempResult);
            break;
        }
        // SDS.
        case 51:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned short tempEndCount = readNumberFromSram(tempLength, argumentAddressList[2]);
            unsigned short tempCount = 0;
            unsigned short index = 0;
            if (tempEndCount)
            {
                startReadingFromSram(argumentAddressList[0]);
                while (true)
                {
                    byte tempCharacter = receiveByteFromSram();
                    if (tempCharacter == ' ')
                    {
                        tempCount += 1;
                    }
                    if (tempCharacter == 0)
                    {
                        break;
                    }
                    index += 1;
                    if (tempCount > tempEndCount - 1)
                    {
                        break;
                    }
                }
                stopConnectionToSram();
            }
            unsigned short tempOffset = 0;
            while (true)
            {
                byte tempCharacter = readNumberFromSram(1, argumentAddressList[0] + index);
                if (tempCharacter == ' ')
                {
                    tempCharacter = 0;
                }
                writeNumberToSram(argumentAddressList[3] + tempOffset, 1, tempCharacter);
                if (tempCharacter == 0)
                {
                    break;
                }
                index += 1;
                tempOffset += 1;
            }
            break;
        }
        // RAN.
        case 52:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned long tempResult = rand() % tempNumber;
            writeNumberToSram(argumentAddressList[2], tempLength, tempResult);
            break;
        }
        // SIN.
        case 53:
        {
            float tempNumber = readFloatFromSram(argumentAddressList[0]);
            float tempResult = sinf(tempNumber);
            writeFloatToSram(argumentAddressList[1], tempResult);
            break;
        }
        // TWR.
        case 54:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            setTimeMilliseconds(tempNumber);
            break;
        }
        // TRD.
        case 55:
        {
            unsigned long tempNumber = getTimeMilliseconds();
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            writeNumberToSram(argumentAddressList[1], tempLength, tempNumber);
            break;
        }
        // TSL.
        case 56:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned long tempNumber = readNumberFromSram(tempLength, argumentAddressList[1]);
            sleepMilliseconds(tempNumber);
            break;
        }
        // DCL.
        case 57:
        {
            clearVideoBuffer();
            break;
        }
        // DWR.
        case 58:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short tempPosX = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned short tempPosY = readNumberFromSram(tempLength, argumentAddressList[2]);
            startReadingFromSram(argumentAddressList[3]);
            while (true)
            {
                byte tempCharacter = receiveByteFromSram();
                if (tempCharacter == 0)
                {
                    break;
                }
                if (tempCharacter == 10)
                {
                    tempPosX = 0;
                    tempPosY += 1;
                } else {
                    unsigned short index = tempPosX + tempPosY * VIDEO_CHARACTER_WIDTH;
                    videoBuffer[index] = tempCharacter;
                    tempPosX += 1;
                    if (tempPosX >= VIDEO_CHARACTER_WIDTH)
                    {
                        tempPosX = 0;
                        tempPosY += 1;
                    }
                }
            }
            stopConnectionToSram();
            break;
        }
        // DRD.
        case 59:
        {
            unsigned short tempLength1 = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short tempPosX = readNumberFromSram(tempLength1, argumentAddressList[1]);
            unsigned short tempPosY = readNumberFromSram(tempLength1, argumentAddressList[2]);
            unsigned short tempLength2 = readNumberFromSram(addressSize, argumentAddressList[3]);
            unsigned short tempSize = readNumberFromSram(tempLength2, argumentAddressList[4]);
            unsigned short index = tempPosX + tempPosY * VIDEO_CHARACTER_WIDTH;
            unsigned short tempOffset = 0;
            startWritingToSram(argumentAddressList[5]);
            while (tempOffset < tempSize)
            {
                byte tempCharacter = videoBuffer[index + tempOffset];
                sendByteToSram(tempCharacter);
                tempOffset += 1;
            }
            sendByteToSram(0);
            stopConnectionToSram();
            break;
        }
        // CKT.
        case 60:
        {
            unsigned short tempKey = readNumberFromSram(2, argumentAddressList[0]);
            byte tempShiftKeyIsHeld = readNumberFromSram(1, argumentAddressList[1]);
            byte tempBuffer[2];
            tempBuffer[0] = convertKeyToCharacter(tempKey, tempShiftKeyIsHeld);
            tempBuffer[1] = 0;
            writeTextToSram(argumentAddressList[2], tempBuffer);
            break;
        }
        // CTK.
        case 61:
        {
            byte tempCharacter = readNumberFromSram(1, argumentAddressList[0]);
            unsigned short tempResult = convertCharacterToKey(tempCharacter);
            writeNumberToSram(argumentAddressList[1], 2, tempResult);
            break;
        }
        // KIP.
        case 62:
        {
            unsigned short tempKey = readNumberFromSram(2, argumentAddressList[0]);
            byte tempResult = keyIsHeld(tempKey);
            writeNumberToSram(argumentAddressList[1], 1, tempResult);
            break;
        }
        // PKY.
        case 63:
        {
            runtimeMode = 1;
            promptDestinationAddress = argumentAddressList[0];
            break;
        }
        // PTX.
        case 64:
        {
            runtimeMode = 2;
            promptDestinationAddress = argumentAddressList[2];
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short tempPosY = readNumberFromSram(tempLength, argumentAddressList[1]);
            unsigned short index = tempPosY * VIDEO_CHARACTER_WIDTH;
            startReadingFromSram(argumentAddressList[2]);
            promptIndex = 0;
            while (true)
            {
                byte tempCharacter = receiveByteFromSram();
                if (tempCharacter == 0)
                {
                    break;
                }
                videoBuffer[index] = tempCharacter;
                index += 1;
                promptIndex += 1;
            }
            stopConnectionToSram();
            videoBuffer[index] = PROMPT_TEXT_CURSOR_CHARACTER;
            promptVideoBufferIndex = index;
            break;
        }
        // PSL.
        case 65:
        {
            runtimeMode = 3;
            promptIndex = 0;
            promptSize = readNumberFromSram(addressSize, argumentAddressList[0]);
            promptAmount = readNumberFromSram(promptSize, argumentAddressList[1]);
            promptSourceAddress = argumentAddressList[2];
            promptDestinationAddress = argumentAddressList[3];
            displayPromptSelection();
            break;
        }
        // PFL.
        case 66:
        {
            runtimeMode = 4;
            promptIndex = 0;
            promptSize = readNumberFromSram(addressSize, argumentAddressList[0]);
            promptAmount = getNumberOfFiles();
            promptDestinationAddress = argumentAddressList[1];
            displayPromptFile();
            break;
        }
        // XSM.
        case 67:
        {
            
            break;
        }
        // XDR.
        case 68:
        {
            
            break;
        }
        // XDW.
        case 69:
        {
            
            break;
        }
        // XAR.
        case 70:
        {
            
            break;
        }
        // RNF.
        case 71:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short tempNumber = getNumberOfFiles();
            writeNumberToSram(argumentAddressList[1], tempLength, tempNumber);
            break;
        }
        // RGAI.
        case 72:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short index = readNumberFromSram(tempLength, argumentAddressList[1]);
            long tempAddress = getFileEntryAddressByIndex(index);
            writeNumberToSram(argumentAddressList[2], 4, tempAddress);
            break;
        }
        // RGAN.
        case 73:
        {
            byte tempBuffer[50];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            long tempAddress = getFileEntryAddressByName(tempBuffer);
            writeNumberToSram(argumentAddressList[1], 4, tempAddress);
            break;
        }
        // RGAA.
        case 74:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            long tempAddress2 = getNextFileEntryAddress(tempAddress);
            writeNumberToSram(argumentAddressList[1], 4, tempAddress2);
            break;
        }
        // RCR.
        case 75:
        {
            byte tempBuffer[50];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned long tempSize = readNumberFromSram(tempLength, argumentAddressList[2]);
            long tempAddress = createFile(tempBuffer, tempSize);
            writeNumberToSram(argumentAddressList[3], 4, tempAddress);
            break;
        }
        // RRD.
        case 76:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            readFile(argumentAddressList[1], tempAddress);
            break;
        }
        // RWR.
        case 77:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[1]);
            writeFile(tempAddress, argumentAddressList[0]);
            break;
        }
        // RGN.
        case 78:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            byte tempBuffer[50];
            getFileName(tempBuffer, tempAddress);
            writeTextToSram(argumentAddressList[1], tempBuffer);
            break;
        }
        // RGS.
        case 79:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[1]);
            unsigned long tempSize = getFileSize(tempAddress);
            writeNumberToSram(argumentAddressList[2], tempLength, tempSize);
            break;
        }
        // RDL.
        case 80:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            deleteFile(tempAddress);
            break;
        }
        // RRN.
        case 81:
        {
            long tempAddress = readNumberFromSram(4, argumentAddressList[0]);
            runFile(tempAddress);
            break;
        }
        // QUIT.
        case 82:
        {
            quitCurrentProgram();
            break;
        }
        // CTO:
        case 83:
        {
            byte tempBuffer[20];
            readTextFromSram(tempBuffer, argumentAddressList[0]);
            byte tempResult = convertTextToOpcode(tempBuffer);
            writeNumberToSram(argumentAddressList[1], 1, tempResult);
            break;
        }
        // CIH.
        case 84:
        {
            byte tempNumber = readNumberFromSram(1, argumentAddressList[0]);
            byte tempBuffer[3];
            tempBuffer[0] = convertNumberToHexDigit((tempNumber & 0xF0) >> 4);
            tempBuffer[1] = convertNumberToHexDigit(tempNumber & 0x0F);
            tempBuffer[2] = 0;
            writeTextToSram(argumentAddressList[1], tempBuffer);
            break;
        }
        // CHI.
        case 85:
        {
            byte tempCharacter = readNumberFromSram(1, argumentAddressList[0]);
            byte tempNumber = convertHexDigitToNumber(tempCharacter) << 4;
            tempCharacter = readNumberFromSram(1, argumentAddressList[0] + 1);
            tempNumber += convertHexDigitToNumber(tempCharacter);
            writeNumberToSram(argumentAddressList[1], 1, tempNumber);
            break;
        }
        // ALLO.
        case 86:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            long tempSize = readNumberFromSram(tempLength, argumentAddressList[1]);
            // Format of allocated memory:
            // (next memory address) (last memory address) (memory size) (data at address)
            // Next address will be zero if there is no next memory.
            long tempAddress = HEAP_START_ADDRESS;
            long tempLastAddress = HEAP_START_ADDRESS;
            long tempNextAddress = HEAP_START_ADDRESS;
            while (true)
            {
                long tempSize2 = readNumberFromSram(2, tempNextAddress - (long)2);
                if (tempNextAddress + tempSize2 + (long)6 + tempSize < tempLastAddress - (long)1)
                {
                    tempAddress = tempNextAddress + tempSize2 + (long)6;
                    break;
                }
                tempLastAddress = tempNextAddress;
                tempNextAddress = readNumberFromSram(2, tempLastAddress - (long)6);
                if (tempNextAddress == 0)
                {
                    tempAddress = tempLastAddress - (long)6 - tempSize;
                    break;
                }
            }
            writeNumberToSram(tempLastAddress - 6, 2, tempAddress);
            writeNumberToSram(tempAddress - 2, 2, tempSize);
            writeNumberToSram(tempAddress - 4, 2, tempLastAddress);
            writeNumberToSram(tempAddress - 6, 2, tempNextAddress);
            if (tempNextAddress > 0)
            {
                writeNumberToSram(tempNextAddress - 4, 2, tempAddress);
            }
            writeNumberToSram(argumentAddressList[2], addressSize, tempAddress);
            break;
        }
        // FREE.
        case 87:
        {
            unsigned short tempAddress = readNumberFromSram(addressSize, argumentAddressList[0]);
            unsigned short tempLastAddress = readNumberFromSram(2, tempAddress - 4);
            unsigned short tempNextAddress = readNumberFromSram(2, tempAddress - 6);
            writeNumberToSram(tempLastAddress - 6, 2, tempNextAddress);
            if (tempNextAddress > 0)
            {
                writeNumberToSram(tempNextAddress - 4, 2, tempLastAddress);
            }
            break;
        }
        // DSC.
        case 88:
        {
            byte tempDirection = readNumberFromSram(1, argumentAddressList[0]);
            byte tempBuffer[20];
            readTextFromSram(tempBuffer, argumentAddressList[1]);
            scrollVideoBuffer(tempDirection, tempBuffer);
            break;
        }
        // DDX.
        case 89:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            writeNumberToSram(argumentAddressList[1], tempLength, VIDEO_CHARACTER_WIDTH);
            break;
        }
        // DDY.
        case 90:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            writeNumberToSram(argumentAddressList[1], tempLength, VIDEO_CHARACTER_HEIGHT);
            break;
        }
        // GMLF.
        case 91:
        {
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[0]);
            short tempWidth = readNumberFromSram(tempLength, argumentAddressList[1]);
            short tempHeight = readNumberFromSram(tempLength, argumentAddressList[2]);
            short tempPosY = 0;
            while (tempPosY < tempHeight)
            {
                short tempPosX = 0;
                while (tempPosX < tempWidth)
                {
                    byte tempCellIsAlive = gameOfLifeCellIsAlive(tempPosX, tempPosY, tempWidth, tempHeight, argumentAddressList[3]);
                    byte tempLiveNeighborCount = 0;
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX - 1, tempPosY - 1, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX, tempPosY - 1, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX + 1, tempPosY - 1, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX - 1, tempPosY, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX + 1, tempPosY, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX - 1, tempPosY + 1, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX, tempPosY + 1, tempWidth, tempHeight, argumentAddressList[3]);
                    tempLiveNeighborCount += gameOfLifeCellIsAlive(tempPosX + 1, tempPosY + 1, tempWidth, tempHeight, argumentAddressList[3]);
                    byte tempResult;
                    if (tempCellIsAlive)
                    {
                        if (tempLiveNeighborCount == 2 || tempLiveNeighborCount == 3)
                        {
                            tempResult = GAME_OF_LIFE_LIVE_CELL;
                        } else {
                            tempResult = GAME_OF_LIFE_DEAD_CELL;
                        }
                    } else {
                        if (tempLiveNeighborCount == 3)
                        {
                            tempResult = GAME_OF_LIFE_LIVE_CELL;
                        } else {
                            tempResult = GAME_OF_LIFE_DEAD_CELL;
                        }
                    }
                    writeNumberToSram(argumentAddressList[4] + tempPosX + tempPosY * tempWidth, 1, tempResult);
                    tempPosX += 1;
                }
                tempPosY += 1;
            }
            break;
        }
        // MNDL.
        case 92:
        {
            float tempRealPart = readFloatFromSram(argumentAddressList[0]);
            float tempImaginaryPart = readFloatFromSram(argumentAddressList[1]);
            float tempOffset = readFloatFromSram(argumentAddressList[2]);
            unsigned short tempLength = readNumberFromSram(addressSize, argumentAddressList[3]);
            long tempCount = readNumberFromSram(tempLength, argumentAddressList[4]);
            byte tempCharacter = MANDELBROT_START_CHARACTER;
            tempCharacter += computeMandelbrotPoint(tempRealPart, tempImaginaryPart, tempCount);
            tempCharacter += computeMandelbrotPoint(tempRealPart + tempOffset, tempImaginaryPart, tempCount) << 1;
            tempCharacter += computeMandelbrotPoint(tempRealPart, tempImaginaryPart + tempOffset, tempCount) << 2;
            tempCharacter += computeMandelbrotPoint(tempRealPart + tempOffset, tempImaginaryPart + tempOffset, tempCount) << 3;
            byte tempBuffer[2];
            tempBuffer[0] = tempCharacter;
            tempBuffer[1] = 0;
            writeTextToSram(argumentAddressList[5], tempBuffer);
            break;
        }
        // POW.
        case 93:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = powf(tempNumber1, tempNumber2);
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        // LOG.
        case 94:
        {
            float tempNumber1 = readFloatFromSram(argumentAddressList[0]);
            float tempNumber2 = readFloatFromSram(argumentAddressList[1]);
            float tempResult = logf(tempNumber1) / logf(tempNumber2);
            writeFloatToSram(argumentAddressList[2], tempResult);
            break;
        }
        default:
        {
            break;
        }
    }
}

void tick(int value)
{
    unsigned short tempKey = getNextKeyPress();
    if (isSleeping)
    {
        if (getTimeMilliseconds() > wakeTimeMilliseconds)
        {
            isSleeping = false;
        }
    } else {
        switch (bootStage)
        {
            case 0:
            {
                sDDirectory = opendir((char *)SD_CARD_DIRECTORY_PATH);
                if (sDDirectory != NULL)
                {
                    drawTextFromProgMem(0, 0, BOOT_MESSAGE_2);
                    drawTextFromProgMem(0, 1, BOOT_MESSAGE_4);
                    drawTextFromProgMem(0, 2, BOOT_MESSAGE_5);
                    sleepMilliseconds(500);
                    bootStage += 1;
                } else {
                    drawTextFromProgMem(0, 0, BOOT_MESSAGE_1);
                    bootStage = 255;
                }
                break;
            }
            case 1:
            {
                long index = 0;
                while (index < sizeof(fileNameList))
                {
                    fileNameList[index] = 0;
                    index += MAXIMUM_FILE_NAME_LENGTH;
                }
                numberOfFiles = 0;
                index = 0;
                rewinddir(sDDirectory);
                struct dirent *tempEntry;
                while (true)
                {
                    tempEntry = readdir(sDDirectory);
                    if (tempEntry == NULL)
                    {
                        break;
                    }
                    copyText(fileNameList + index, (byte *)(tempEntry->d_name));
                    numberOfFiles += 1;
                    index += MAXIMUM_FILE_NAME_LENGTH;
                }
                byte tempName[50];
                readTextFromProgMem(tempName, START_FILE_NAME);
                long tempAddress = getFileEntryAddressByName(tempName);
                if (tempAddress != -1)
                {
                    runFile(tempAddress);
                    bootStage = 2;
                } else {
                    drawTextFromProgMem(0, 3, BOOT_MESSAGE_6);
                    bootStage = 255;
                }
                break;
            }
            case 2:
            {
                if (tempKey == 0x76)
                {
                    quitCurrentProgram();
                }
                switch (runtimeMode)
                {
                    case 0:
                    {
                        byte tempCount = 0;
                        while (tempCount < 100)
                        {
                            processNextCommand();
                            if (runtimeMode != 0 || isSleeping)
                            {
                                break;
                            }
                            tempCount += 1;
                        }
                        break;
                    }
                    case 1:
                    {
                        if (tempKey)
                        {
                            writeNumberToSram(promptDestinationAddress, 2, tempKey);
                            runtimeMode = 0;
                        }
                        break;
                    }
                    case 2:
                    {
                        byte shiftKeyIsHeld = (keyIsHeld(0x0012) || keyIsHeld(0x0059));
                        byte tempCharacter = convertKeyToCharacter(tempKey, shiftKeyIsHeld);
                        if (tempCharacter != 0)
                        {
                            writeNumberToSram(promptDestinationAddress + promptIndex, 1, tempCharacter);
                            promptIndex += 1;
                            writeNumberToSram(promptDestinationAddress + promptIndex, 1, 0);
                            videoBuffer[promptVideoBufferIndex] = tempCharacter;
                            promptVideoBufferIndex += 1;
                            videoBuffer[promptVideoBufferIndex] = PROMPT_TEXT_CURSOR_CHARACTER;
                        }
                        // Backspace.
                        if (tempKey == 0x66 && promptIndex)
                        {
                            promptIndex -= 1;
                            writeNumberToSram(promptDestinationAddress + promptIndex, 1, 0);
                            videoBuffer[promptVideoBufferIndex] = ' ';
                            promptVideoBufferIndex -= 1;
                            videoBuffer[promptVideoBufferIndex] = PROMPT_TEXT_CURSOR_CHARACTER;
                        }
                        // Return.
                        if (tempKey == 0x5A)
                        {
                            runtimeMode = 0;
                        }
                        break;
                    }
                    case 3:
                    case 4:
                    {
                        byte tempLastPosY = promptIndex - promptIndex2;
                        // Up.
                        if (tempKey == 0xE075 && promptIndex)
                        {
                            promptIndex -= 1;
                        }
                        // Down.
                        if (tempKey == 0xE072 && promptIndex < promptAmount - 1)
                        {
                            promptIndex += 1;
                        }
                        // B.
                        if (tempKey == 0x0032)
                        {
                            if (promptIndex > VIDEO_CHARACTER_HEIGHT)
                            {
                                promptIndex -= VIDEO_CHARACTER_HEIGHT;
                            } else {
                                promptIndex = 0;
                            }
                        }
                        // N.
                        if (tempKey == 0x0031)
                        {
                            
                            if (promptAmount - promptIndex > VIDEO_CHARACTER_HEIGHT)
                            {
                                promptIndex += VIDEO_CHARACTER_HEIGHT;
                            } else {
                                promptIndex = promptAmount - 1;
                            }
                        }
                        if (promptIndex < promptIndex2 || promptIndex > promptIndex2 + VIDEO_CHARACTER_HEIGHT - 1)
                        {
                            if (runtimeMode == 4)
                            {
                                displayPromptFile();
                            } else {
                                displayPromptSelection();
                            }
                        } else {
                            byte tempPosY = promptIndex - promptIndex2;
                            if (tempPosY != tempLastPosY)
                            {
                                videoBuffer[tempLastPosY * VIDEO_CHARACTER_WIDTH] = ' ';
                                videoBuffer[tempPosY * VIDEO_CHARACTER_WIDTH] = PROMPT_SELECTION_CURSOR_CHARACTER;
                            }
                        }
                        // Return.
                        if (tempKey == 0x5A)
                        {
                            writeNumberToSram(promptDestinationAddress, promptSize, promptIndex);
                            runtimeMode = 0;
                        }
                        break;
                    }
                    default:
                    {
                        break;
                    }
                }
            }
            default:
            {
                break;
            }
        }
    }
    
    display();
    timeMilliseconds += FRAME_DELAY;
    glutTimerFunc(FRAME_DELAY, tick, 0);
}

int main(int argc, char** argv)
{
    clearVideoBuffer();
    
    glutInit(&argc, argv);
    
    glutInitWindowSize(VIEWPORT_WIDTH, VIEWPORT_HEIGHT);
    
    glutInitDisplayMode(GLUT_RGB | GLUT_SINGLE);
    
    glutCreateWindow("DUO Light Emulator");
    
    glutDisplayFunc(display);
    
    glutReshapeFunc(reshape);
    
    glutKeyboardFunc(keyboardEvent);
    glutKeyboardUpFunc(keyboardUpEvent);
    glutSpecialFunc(specialEvent);
    glutSpecialUpFunc(specialUpEvent);
    
    tick(0);
    
    glClearColor(0,0,0,1);
    
    glutMainLoop();
    
    return 0;
}



